package core

import (
	"os"

	"github.com/grufwub/go-errors"
)

// FileContent provides an interface for caching, rendering and getting cached contents of a file
type FileContent interface {
	Load(*Path, *os.File) *errors.Error
	WriteToClient(*Client, *Path) *errors.Error
	Clear()
}

// generatedFileContents is a simple FileContent implementation for holding onto a generated (virtual) file contents
type generatedFileContent struct {
	content []byte
}

// ReadAllFrom does nothing for generated content
func (fc *generatedFileContent) Load(p *Path, file *os.File) *errors.Error { return nil }

// WriteAllTo writes the generated FileContent to client
func (fc *generatedFileContent) WriteToClient(client *Client, p *Path) *errors.Error {
	return client.Conn().Write(fc.content)
}

// Clear does nothing
func (fc *generatedFileContent) Clear() {}

// RegularFileContent is the simplest implementation of core.FileContents for regular files
type RegularFileContent struct {
	content []byte
}

// Load takes an open FD and loads the file contents into FileContents memory
func (fc *RegularFileContent) Load(p *Path, file *os.File) *errors.Error {
	var err *errors.Error
	fc.content, err = ReadFile(file)
	return err
}

// WriteToClient writes the current contents of FileContents to the client
func (fc *RegularFileContent) WriteToClient(client *Client, p *Path) *errors.Error {
	return client.Conn().Write(fc.content)
}

// Clear empties currently cached FileContents memory
func (fc *RegularFileContent) Clear() {
	fc.content = nil
}
