package core

import (
	"io"
	"os"
	"sort"

	"github.com/grufwub/go-errors"
	"github.com/grufwub/go-filecache"
)

// OpenFile opens a file for reading (read-only, world-readable)
func OpenFile(p *Path) (*os.File, *errors.Error) {
	file, err := os.OpenFile(p.Absolute(), os.O_RDONLY, 0444)
	return file, errors.WrapIfError(FileOpenErr, err)
}

// StatFile performs a file stat on a file at path
func StatFile(p *Path) (os.FileInfo, *errors.Error) {
	stat, err := os.Stat(p.Absolute())
	return stat, errors.WrapIfError(FileStatErr, err)
}

// StatForFile gets the state from a given file
func StatForFile(file *os.File) (os.FileInfo, *errors.Error) {
	stat, err := file.Stat()
	return stat, errors.WrapIfError(FileStatErr, err)
}

// ReadFile reads a supplied file descriptor into a return byte slice, or error
func ReadFile(file *os.File) ([]byte, *errors.Error) {
	// Return slice
	ret := make([]byte, 0)

	// Get read buffers, defer putting back
	br := fileBufferedReaderPool.Get(file)
	defer fileBufferedReaderPool.Put(br)

	// Read through file until null bytes / error
	for {
		// Read line
		line, err := br.ReadBytes('\n')
		if err != nil {
			if err == io.EOF {
				// EOF, add current to return slice and
				// break-out. WIll not have hit delim
				ret = append(ret, line...)
				break
			} else {
				// Bad error, return
				return nil, errors.WrapError(FileReadErr, err)
			}
		}

		// Add current line to return slice
		ret = append(ret, line...)
	}

	// Return!
	return ret, nil
}

// ScanFile scans a supplied file at file descriptor, using iterator function
func ScanFile(file *os.File, iterator func(string) bool) *errors.Error {
	// Get read buffer, defer putting back
	br := fileBufferedReaderPool.Get(file)
	defer fileBufferedReaderPool.Put(br)

	// Iterate through file!
	for {
		// Read a line
		line, err := br.ReadString('\n')
		if err != nil {
			if err == io.EOF {
				// Reached end of file, perform final iteration
				// and break-out. Will not have hit delim
				iterator(line)
				break
			} else {
				// Bad error, return
				return errors.WrapError(FileReadErr, err)
			}
		}

		// Run scan iterator on this line, breaking out if requested,
		// skipping final byte which is '\n'
		if !iterator(line[:len(line)-1]) {
			break
		}
	}

	// Return no errors :)
	return nil
}

// ScanDirectory reads the contents of a directory and performs the iterator function on each os.FileInfo entry returned
func ScanDirectory(dir *os.File, p *Path, iterator func(os.FileInfo, *Path)) *errors.Error {
	dirList, err := dir.Readdir(-1)
	if err != nil {
		return errors.WrapError(DirectoryReadErr, err)
	}

	// Sort by name
	sort.Sort(byName(dirList))

	// Walk through the directory list using supplied iterator function
	for _, info := range dirList {
		// Make new Path object
		fp := p.JoinPath(info.Name())

		// Skip restricted files
		if IsRestrictedPath(fp) || IsHiddenPath(fp) || WithinCGIDir(fp) {
			continue
		}

		// Perform iterator
		iterator(info, p.JoinPath(info.Name()))
	}

	return nil
}

// AddGeneratedFile adds a generated file content byte slice to the file cache, with supplied path as the key
func AddGeneratedFile(p *Path, b []byte) {
	// Get write lock, defer unlock
	FileCache.Lock()
	defer FileCache.Unlock()

	// Wrap contents in File
	file := filecache.NewFile(p.Absolute(), false, &generatedFileContent{b})

	// Add to cache!
	FileCache.Put(file)
}
